#include <stdio.h>
#include <ctype.h>
#include "uart.h"
#include "hexdump.h"

#define AL_LE_N A5
#define AH_LE_N A4
#define DL_LE_N A3
#define DH_LE_N A2
#define DL_OE_N A1
#define DH_OE_N A0

#define OE_N 10
#define CE_N 11
#define PGM_N 12
#define VPP_EN 13

uint8_t buf[0xFF];

/* databus manipulation routines */
void db_take()
{
    DDRD |= 0xFC;
    DDRB |= 0x03;
}

void db_release()
{
    DDRD &= ~0xFC;
    DDRB &= ~0x03;
}

uint8_t db_read()
{
    return (PIND >> 2) | (PINB << 6);
}

void db_write(uint8_t data)
{
    PORTD &= ~0xFC;
    PORTD |= (data << 2);
    PORTB &= ~0x03;
    PORTB |= (data >> 6);
}

/* high level routines */
uint16_t read16(uint16_t addr)
{
    uint8_t rl, rh;
    
    /* release addr latches */
    digitalWrite(AL_LE_N, HIGH);
    digitalWrite(AH_LE_N, HIGH);
    delayMicroseconds(1);

    /* write address */
    db_take();
    db_write(lowByte(addr));
    delayMicroseconds(1);
    digitalWrite(AL_LE_N, LOW);
    delayMicroseconds(1);
    db_write(highByte(addr));
    delayMicroseconds(1);
    digitalWrite(AH_LE_N, LOW);
    delayMicroseconds(1);
    db_release();
    
    /* wait for data */
    delayMicroseconds(1);

    /* read data */
    digitalWrite(DL_OE_N, LOW);
    delayMicroseconds(1);
    rl = db_read();
    digitalWrite(DL_OE_N, HIGH);
    digitalWrite(DH_OE_N, LOW);
    delayMicroseconds(1);
    rh = db_read();
    digitalWrite(DH_OE_N, HIGH);

    return word(rh, rl);
}

uint16_t program16(uint16_t addr, uint16_t data)
{
    uint8_t rl, rh;
    
    /* release addr latches */
    digitalWrite(AL_LE_N, HIGH);
    digitalWrite(AH_LE_N, HIGH);
    delayMicroseconds(1);

    /* write address and data */
    db_take();
    /* address */ 
    db_write(lowByte(addr));
    delayMicroseconds(1);
    digitalWrite(AL_LE_N, LOW);
    delayMicroseconds(1);
    db_write(highByte(addr));
    delayMicroseconds(1);
    digitalWrite(AH_LE_N, LOW);
    delayMicroseconds(1);
    /* data */
    db_write(lowByte(data));
    delayMicroseconds(1);
    digitalWrite(DL_LE_N, LOW);
    db_write(highByte(data));
    delayMicroseconds(1);
    digitalWrite(DH_LE_N, LOW);
    delayMicroseconds(1);
    db_release();
    
    /* wait for address and data to settle */
    delayMicroseconds(2);
    
    /* pulse PGM_N for 100us */
    digitalWrite(PGM_N, LOW);
    delayMicroseconds(100);
    digitalWrite(PGM_N, HIGH);

    /* keep address and data as required */
    delayMicroseconds(2);

    /* release data */
    digitalWrite(DL_LE_N, HIGH);
    digitalWrite(DH_LE_N, HIGH);
    delayMicroseconds(1);

    /* read data back */
    digitalWrite(OE_N, LOW);
    digitalWrite(DL_OE_N, LOW);
    delayMicroseconds(2);
    rl = db_read();
    digitalWrite(DL_OE_N, HIGH);
    delayMicroseconds(1);
    digitalWrite(DH_OE_N, LOW);
    delayMicroseconds(1);
    rh = db_read();
    digitalWrite(DH_OE_N, HIGH);
    digitalWrite(OE_N, HIGH);

    return word(rh, rl);
}

void setup() {
    db_release();
    digitalWrite(AL_LE_N, HIGH);
    digitalWrite(AH_LE_N, HIGH);
    digitalWrite(DL_LE_N, HIGH);
    digitalWrite(DH_LE_N, HIGH);
    digitalWrite(DL_OE_N, HIGH);
    digitalWrite(DH_OE_N, HIGH);
    pinMode(AL_LE_N, OUTPUT); 
    pinMode(AH_LE_N, OUTPUT);
    pinMode(DL_LE_N, OUTPUT);
    pinMode(DH_LE_N, OUTPUT);
    pinMode(DL_OE_N, OUTPUT);
    pinMode(DH_OE_N, OUTPUT);
    
    digitalWrite(OE_N, HIGH);
    digitalWrite(CE_N, HIGH);
    digitalWrite(PGM_N, HIGH);
    pinMode(OE_N, OUTPUT);
    pinMode(CE_N, OUTPUT);
    pinMode(PGM_N, OUTPUT);

    uart_init();
}

void loop() {
    printf("\r\nflashprog>");
    char cmd = getchar();
    if(cmd == 'T')
    {
        printf("\r\nTEST1:\r\nconnect addr to data and press any key");
        getchar();
        for(int i = 0; i < 16; i++)
        {
            printf("\r\n%04X: %04X", 1 << i, read16(1 << i)); 
        }
        printf("\r\n%04X: %04X", 0x5555, read16(0x5555));  
        printf("\r\n%04X: %04X", 0xAAAA, read16(0xAAAA));  
        
        printf("\r\nTEST2:\r\nconnect data low byte to data high byte");
        getchar();
        for(int i = 0; i < 8; i++)
        {
            uint8_t in = 1 << i, out;
            db_take();
            db_write(in);
            delayMicroseconds(1);
            digitalWrite(DL_LE_N, LOW);
            delayMicroseconds(1);
            db_release();
            digitalWrite(DH_OE_N, LOW);
            delayMicroseconds(1);
            out = db_read();
            digitalWrite(DL_LE_N, HIGH);
            digitalWrite(DH_OE_N, HIGH);

            printf("\r\n%02X: %02X", in, out); 
        }
        for(int i = 0; i < 8; i++)
        {
            uint8_t in = 1 << i, out;
            db_take();
            db_write(in);
            delayMicroseconds(1);
            digitalWrite(DH_LE_N, LOW);
            delayMicroseconds(1);
            db_release();
            digitalWrite(DL_OE_N, LOW);
            delayMicroseconds(1);
            out = db_read();
            digitalWrite(DH_LE_N, HIGH);
            digitalWrite(DL_OE_N, HIGH);

            printf("\r\n%02X: %02X", in, out); 
        }
        
        printf("\r\nTEST3:\r\nconnect OE, CE and PGM to data0..2 and press any key");
        getchar();
        digitalWrite(DL_OE_N, LOW);
        delayMicroseconds(1);
        digitalWrite(OE_N, HIGH); digitalWrite(CE_N, LOW); digitalWrite(PGM_N, LOW);
        printf("\r\n01 %02X", db_read() & 0x07);
        digitalWrite(OE_N, LOW); digitalWrite(CE_N, HIGH); digitalWrite(PGM_N, LOW);
        printf("\r\n02 %02X", db_read() & 0x07);
        digitalWrite(OE_N, LOW); digitalWrite(CE_N, LOW); digitalWrite(PGM_N, HIGH);
        printf("\r\n04 %02X", db_read() & 0x07);
        digitalWrite(OE_N, HIGH); digitalWrite(CE_N, LOW); digitalWrite(PGM_N, HIGH);
        printf("\r\n05 %02X", db_read() & 0x07);
        digitalWrite(OE_N, LOW); digitalWrite(CE_N, LOW); digitalWrite(PGM_N, LOW);
        printf("\r\n00 %02X", db_read() & 0x07);
        digitalWrite(OE_N, HIGH); digitalWrite(CE_N, HIGH); digitalWrite(PGM_N, HIGH);
        printf("\r\n07 %02X", db_read() & 0x07);
    }
    else if(cmd == 'R')
    {
        digitalWrite(CE_N, LOW);
        digitalWrite(OE_N, LOW);
        delayMicroseconds(1);
        for(uint32_t i = 0; i < 100; i++)
        {
            uint16_t w;
            uint8_t lsb, msb;
            
            //Serial.write(read16(i));
            w = read16(i);
            lsb = lowByte(w);
            msb = highByte(w);
            printf("%02X", msb);
            //if(isprint(msb)) printf("%c", msb);
            printf(" ");
            printf("%02X", lsb);
            //if(isprint(lsb)) printf("%c", lsb);
            printf(" ");
        }
        digitalWrite(CE_N, HIGH);
        digitalWrite(OE_N, HIGH);
    }
    else if(cmd == 'C')
    {
        printf("\r\nchecking EPROM for non-blank words...\r\n");
        digitalWrite(CE_N, LOW);
        digitalWrite(OE_N, LOW);
        delayMicroseconds(1);
        uint32_t non_blank = 0;
        for(uint32_t i = 0; i < 65536; i++)
        {
            uint16_t w;
            
            w = read16(i);
            if(w != 0xFFFF)
                non_blank++;
            if(i % 1024 == 0)
                printf(".");
        }
        printf("\r\ncheck done: %ld non-blank words", non_blank);
        digitalWrite(CE_N, HIGH);
        digitalWrite(OE_N, HIGH);
    }
    else if(cmd == 'D')
    {
        // TODO handle len check, endianess
        uint16_t addr, len = 0x80;
        char subcmd = 'N';

        printf("\r\nstart address?\r\n");
        scanf("%X", &addr);
        
        while(subcmd == 'N')
        {
            digitalWrite(CE_N, LOW);
            digitalWrite(OE_N, LOW);
            delayMicroseconds(1);
            for(uint16_t i = 0; i < len; i++)
            {
                int w = read16(addr+i);
                buf[2*i] = highByte(w);
                buf[2*i+1] = lowByte(w);
            }
            digitalWrite(CE_N, HIGH);
            digitalWrite(OE_N, HIGH);
            hexdump(buf, 2*len, addr);

            printf("N/Q ?");
            do
            {
                subcmd = getchar();
            }
            while(subcmd != 'N' && subcmd != 'Q');
            addr += len;
            printf("\r\n");
        }
    }
    else if(cmd == 'P')
    {
        uint8_t data[] = "Coucou from FLASHPROG :)";
        
        digitalWrite(CE_N, LOW);
        delayMicroseconds(1);
        for(uint16_t i = 0; i < 12; i++)
        {
            uint16_t w = word(data[2*i],data[2*i+1]);
            printf("%04X: %04X\r\n", w, program16(i, w));
        }
        digitalWrite(CE_N, HIGH);
    }
    else
    {
        printf("\r\nT - self test");    
    }

	delay(50);
}
