library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_unsigned.all;
use ieee.numeric_std.all;
 
entity bfm is
end entity;
 
architecture TB of bfm is

	component ram

        generic (
            DATA_BITS : natural;
            ADDR_BITS : natural;
            DATA_FILE : string;
            TIME_RC : time;
            TIME_HZ : time
        ); 

        port (
             addr    : in std_logic_vector (ADDR_BITS-1 downto 0);
             wr_n	 : in std_logic;
             oe		 : in std_logic;
             data    : inout std_logic_vector (DATA_BITS-1 downto 0)
         );

	end component;
	
    component counter is
		generic (
            WIDTH : integer := 8;
            DELAY : time
        ); 
		port (
			output   : out std_logic_vector (WIDTH-1 downto 0);
			zero	 : out std_logic;
			input    : in std_logic_vector (WIDTH-1 downto 0);
			up		 : in std_logic;
			down	 : in std_logic;
			reset	 : in std_logic;
			preset	 : in std_logic
		);
	end component;
	

    signal addr_bus  : std_logic_vector (7 downto 0) := (others => 'Z');
    signal data_bus  : std_logic_vector (7 downto 0) := (others => 'Z');

    signal prg_addr  : std_logic_vector (7 downto 0) := (others => 'Z');
    signal prg_data  : std_logic_vector (15 downto 0) := (others => 'Z');

    signal ctrl_data : std_logic_vector (15 downto 0) := (others => 'Z');
    signal ctrl_addr : std_logic_vector (5 downto 0) := (others => 'Z');
        
    signal pc_output : std_logic_vector (7 downto 0) := (others => 'Z');
    signal pc_zero : std_logic := 'Z';
    signal pc_input : std_logic_vector (7 downto 0) := (others => 'Z');
    signal pc_up : std_logic := 'Z';
    signal pc_reset : std_logic := 'Z';
    signal pc_preset : std_logic := 'Z';
        
    signal subclk_output : std_logic_vector (1 downto 0) := (others => 'Z');
    signal subclk_zero  : std_logic := 'Z';
    signal subclk_reset : std_logic := 'Z';
    
    signal ram_wr_n  : std_logic := 'Z';
    signal ram_oe  : std_logic := 'Z';
    
    signal clk  : std_logic := 'Z';
    signal reset  : std_logic := 'Z';

begin
    
    -- working RAM
	bfm_ram : ram
    generic map (
		DATA_BITS => 8,
		ADDR_BITS => 8,
		DATA_FILE => "test/bfm_ram.dat",
        TIME_RC => 15 ns,
        TIME_HZ => 8 ns
	)
	port map (
        addr => addr_bus,
        wr_n => ram_wr_n,
        oe => ram_oe,
        data => data_bus
    );
	
    -- program ROM
    prg_rom : ram
    generic map (
		DATA_BITS => 16,
		ADDR_BITS => 8,
		DATA_FILE => "test/bfm_prg_rom.dat",
        TIME_RC => 15 ns,
        TIME_HZ => 8 ns
	)
	port map (
        addr => prg_addr,
        wr_n => '1',
        oe => '1',
        data => prg_data
    );
    
    -- control ROM
    ctrl_rom : ram
    generic map (
		DATA_BITS => 16,
		ADDR_BITS => 6,
		DATA_FILE => "test/bfm_ctrl_rom.dat",
        TIME_RC => 15 ns,
        TIME_HZ => 8 ns
	)
	port map (
        addr => ctrl_addr,
        wr_n => '1',
        oe => '1',
        data => ctrl_data
    );

    -- PC register
	pc : counter
    generic map (
        WIDTH => 8,
        DELAY => 5 ns
    )
    port map (
        output => pc_output,
        zero => pc_zero,
        input => pc_input,
        up => pc_up,
        down => '0',
        reset => pc_reset,
        preset => pc_preset
    );
    
    -- clk divider
	subclk : counter
    generic map (
        WIDTH => 2,
        DELAY => 5 ns
    )
    port map (
        output => subclk_output,
        zero => subclk_zero,
        input => (others => 'Z'),
        up => clk,
        down => '0',
        reset => subclk_reset,
        preset => '0'
    );
    
    -- 10 MHz clock
    process
    begin
        --data <= (others => 'Z');
        clk <= '0';
        wait for 50 ns;
        clk <= '1';
        wait for 50 ns;
    end process;
    
    -- ctrl signals
    ctrl_addr(1 downto 0) <= subclk_output;
    ctrl_addr(5 downto 2) <= prg_data(15 downto 12);
    pc_up <= ctrl_data(0);
    
    -- reset signal
    pc_reset <= reset;
    subclk_reset <= reset;

    -- launch everything by pulsing reset
    process
    begin
        reset <= '0';
        wait for 10 ns;
        reset <= '1';
        wait for 50 ns;
        reset <= '0';
        wait for 10 us;
    end process;

end;
