library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_unsigned.all;
use ieee.numeric_std.all;
 
entity bfm2 is
end entity;
 
architecture TB of bfm2 is

	component ram

        generic (
            DATA_BITS : natural;
            ADDR_BITS : natural;
            DATA_FILE : string;
            TIME_RC : time;
            TIME_HZ : time
        ); 

        port (
             addr    : in std_logic_vector (ADDR_BITS-1 downto 0);
             wr_n	 : in std_logic;
             oe		 : in std_logic;
             data    : inout std_logic_vector (DATA_BITS-1 downto 0)
         );

	end component;

    component counter_74590

        generic (
            DELAY : time
        ); 

        port (
            output   : out std_logic_vector (7 downto 0);
            oe_n	 : in std_logic;
            rclk	 : in std_logic;
            cclk	 : in std_logic;
            ccken_n	 : in std_logic;
            cclr_n	 : in std_logic;
            rco_n    : out std_logic
        );

    end component;

    component adder

        generic (
            WIDTH : integer;
            DELAY : time
        ); 

        port (
            a        : in std_logic_vector (WIDTH-1 downto 0);
            b        : in std_logic_vector (WIDTH-1 downto 0);
            s        : out std_logic_vector (WIDTH-1 downto 0);
            c_out    : out std_logic;
            c_in     : in std_logic
        );

    end component;

    component dff

        generic (
            WIDTH : integer;
            DELAY : time
        ); 

        port (
            output   : out std_logic_vector (WIDTH-1 downto 0);
            input    : in std_logic_vector (WIDTH-1 downto 0);
            oe_n	 : in std_logic;
            clk	     : in std_logic
        );

    end component;

    component buff

        generic (
            WIDTH : integer;
            DELAY : time
        ); 

        port (
            output   : out std_logic_vector (WIDTH-1 downto 0);
            input    : in std_logic_vector (WIDTH-1 downto 0);
            oe_n	 : in std_logic
        );

    end component;

    -- CPU interface
    signal addr_bus  : std_logic_vector (15 downto 0);
    signal data_bus  : std_logic_vector (15 downto 0);
    signal clk  : std_logic;
    signal reset  : std_logic;

    -- ctrl signals
    signal clkdiv_cclr_n : std_logic;

    signal mem_wr_n : std_logic;
    signal mem_oe : std_logic;
    
    signal ir_clk : std_logic;
    
    signal al_oe_n : std_logic;
    signal al_clk : std_logic;

    signal ah_oe_n : std_logic;
    signal ah_clk : std_logic;

    signal ao_oe_n : std_logic;
    
    signal p_oe_n : std_logic;
    signal p_clk : std_logic;
    
    -- carry out
    signal add_c_out : std_logic;

    -- intermediate interconnect signals we don't care about
    signal ctrl_data : std_logic_vector (15 downto 0);
    signal ctrl_addr : std_logic_vector (11 downto 0);
    signal adder_b   : std_logic_vector (15 downto 0);
    signal adder_s   : std_logic_vector (15 downto 0);
	signal clkdiv_output : std_logic_vector (7 downto 0);
    signal clkdiv_rco_n : std_logic;

begin
    
    -- memory (RAM and ROM)
    mem : ram
    generic map (
		DATA_BITS => 16,
		ADDR_BITS => 16,
		DATA_FILE => "test/bfm2_mem.dat",
        TIME_RC => 15 ns,
        TIME_HZ => 8 ns
	)
	port map (
        addr => addr_bus,
        wr_n => mem_wr_n,
        oe => mem_oe,
        data => data_bus
    );
    
    -- control ROM
    ctrl : ram
    generic map (
		DATA_BITS => 16,
		ADDR_BITS => 12,
		DATA_FILE => "test/bfm2_ctrl.dat",
        TIME_RC => 15 ns,
        TIME_HZ => 8 ns
	)
	port map (
        addr => ctrl_addr,
        wr_n => '1',
        oe => '1',
        data => ctrl_data
    );
	
    -- instruction register
    ir : dff
    generic map (
        WIDTH => 8,
        DELAY => 8 ns
    )
    port map (
        output => ctrl_addr(11 downto 4),
        input => data_bus(15 downto 8),
        oe_n => '0',
        clk => ir_clk
    );
    
    -- A register, lower byte
    al : dff
    generic map (
        WIDTH => 8,
        DELAY => 8 ns
    )
    port map (
        output => adder_b(7 downto 0),
        input => adder_s(7 downto 0),
        oe_n => al_oe_n,
        clk => al_clk
    );
    
    -- A register, higher byte
    ah : dff
    generic map (
        WIDTH => 8,
        DELAY => 8 ns
    )
    port map (
        output => adder_b(15 downto 8),
        input => adder_s(15 downto 8),
        oe_n => ah_oe_n,
        clk => ah_clk
    );
    
    -- buffer from A output to data bus
    ao : buff
    generic map (
        WIDTH => 16,
        DELAY => 8 ns
    )
    port map (
        output => data_bus(15 downto 0),
        input => adder_b(15 downto 0),
        oe_n => ao_oe_n
    );
    
    -- P register
    p : dff
    generic map (
        WIDTH => 16,
        DELAY => 8 ns
    )
    port map (
        output => addr_bus,
        input => data_bus,
        oe_n => p_oe_n,
        clk => p_clk
    );
    
    -- adder    
    add : adder
    generic map (
        WIDTH => 16,
        DELAY => 40 ns
    )
    port map (
        a => data_bus,
        b => adder_b,
        s => adder_s,
        c_out => add_c_out,
        c_in => '0'
    );

    -- clock divider
	clkdiv : counter_74590
    generic map (
        DELAY => 5 ns
    )
    port map (
		output => clkdiv_output,
		oe_n   => '0',
		rclk   => clk,
		cclk   => clk,
		ccken_n	=> '0',
		cclr_n => clkdiv_cclr_n,
        rco_n => clkdiv_rco_n 
    );
    
    -- 10 MHz main clock
    process
    begin
        --data <= (others => 'Z');
        clk <= '0';
        wait for 50 ns;
        clk <= '1';
        wait for 50 ns;
    end process;
    
    -- /!\ bus data and addr are pulled low when nobody drives them
    addr_bus <= (others => 'L');
    data_bus <= (0 => 'H', others => 'L') when (ctrl_data(9) = '1') else (others => 'L');
    -- weak drive adder input b to 0
    adder_b <= (others => 'L');

    -- ctrl signals
    ctrl_addr(3 downto 0) <= clkdiv_output(3 downto 0);
    
    -- next_n : low to indicate the end of the instruction
    ir_clk <= (ctrl_data(0) and (not reset)) or clk;
    clkdiv_cclr_n <= (ctrl_data(0) and (not reset));

    mem_oe <= ctrl_data(1);
    mem_wr_n <= ctrl_data(2);
    
    al_oe_n <= ctrl_data(3);
    ah_oe_n <= ctrl_data(3);
    al_clk <= ctrl_data(4) or clk;
    ah_clk <= ctrl_data(5) or clk;
    ao_oe_n <= ctrl_data(6);
    
    p_oe_n <= ctrl_data(7);
    p_clk <= ctrl_data(8);
    
    -- launch everything by pulsing reset
    process
    begin
        reset <= '0';
        wait for 10 ns;
        reset <= '1';
        wait for 50 ns;
        reset <= '0';
        wait for 10 us;
    end process;

end;
