library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_unsigned.all;
use ieee.numeric_std.all;

use work.txt_util.all;

entity bfm3 is
end entity;
 
architecture TB of bfm3 is

	component ram

        generic (
            DATA_BITS : natural;
            ADDR_BITS : natural;
            DATA_FILE : string;
            TIME_RC : time;
            TIME_HZ : time
        ); 

        port (
             addr    : in std_logic_vector (ADDR_BITS-1 downto 0);
             wr_n	 : in std_logic;
             oe		 : in std_logic;
             data    : inout std_logic_vector (DATA_BITS-1 downto 0)
         );

	end component;
    
    -- up/down counter, asynchronous preset :(
    component counter_74191 is

        generic (
            WIDTH : integer;
            DELAY : time
        ); 

        port (
            output   : out std_logic_vector (WIDTH-1 downto 0);
            input    : in std_logic_vector (WIDTH-1 downto 0);
            clk		 : in std_logic;
            cten_n	 : in std_logic;
            du_n	 : in std_logic;
            load_n	 : in std_logic;
            rco_n    : out std_logic;
            minmax   : out std_logic
        );

    end component;

    -- up counter, synchronous preset
    component counter_74163 is

        generic (
            WIDTH : integer;
            DELAY : time
        ); 

        port (
            output   : out std_logic_vector (WIDTH-1 downto 0);
            input    : in std_logic_vector (WIDTH-1 downto 0);
            clk		 : in std_logic;
            cet		 : in std_logic;
            cep		 : in std_logic;
            sr_n	 : in std_logic;
            pe_n     : in std_logic;
            tc	     : out std_logic
        );

    end component;

    component buff

        generic (
            WIDTH : integer;
            DELAY : time
        ); 

        port (
            output   : out std_logic_vector (WIDTH-1 downto 0);
            input    : in std_logic_vector (WIDTH-1 downto 0);
            oe_n	 : in std_logic
        );

    end component;

    -- CPU interface
    signal pm_addr  : std_logic_vector (15 downto 0);
    signal pm_data  : std_logic_vector (15 downto 0);
    signal ram_addr  : std_logic_vector (15 downto 0);
    signal ram_data  : std_logic_vector (7 downto 0);
    signal clk  : std_logic;
    signal reset  : std_logic;

    -- control signals
    signal co_out  : std_logic_vector (15 downto 0);
    signal pm_oe : std_logic;

    signal ram_wr_n : std_logic;
    signal ram_oe : std_logic;
    
    signal pc_sr_n : std_logic;
    signal pc_pe_n : std_logic;
    
    signal p_cten_n : std_logic;
    signal p_load_n : std_logic;
    signal p_du_n : std_logic;
    
    signal a_cten_n : std_logic;
    signal a_load_n : std_logic;
    signal a_du_n : std_logic;
        
    signal ao_oe_n : std_logic;
    
    -- interconnect signals
    signal a_output : std_logic_vector (7 downto 0);
    signal pc_input : std_logic_vector (15 downto 0);

    -- unsused signals
    signal pc_tc : std_logic;
    signal p_rco_n : std_logic;
    signal p_minmax : std_logic;
    signal a_rco_n : std_logic;
    signal a_minmax : std_logic;

begin
    -- PM 
    pm : ram
    generic map (
		DATA_BITS => 16,
		ADDR_BITS => 16,
		DATA_FILE => "test/bfm3_pm.dat",
        TIME_RC => 15 ns,
        TIME_HZ => 8 ns
	)
	port map (
        addr => pm_addr,
        wr_n => '1',
        oe => pm_oe,
        data => pm_data
    );

    -- RAM
    mem : ram
    generic map (
		DATA_BITS => 8,
		ADDR_BITS => 16,
		DATA_FILE => "test/bfm3_ram.dat",
        TIME_RC => 15 ns,
        TIME_HZ => 8 ns
	)
	port map (
        addr => ram_addr,
        wr_n => ram_wr_n,
        oe => ram_oe,
        data => ram_data
    );
    
    -- PC register/counter
    pc : counter_74163
    generic map (
        WIDTH => 16,
        DELAY => 15 ns
    ) 
    port map (
        output   =>     pm_addr,
        input    =>     pc_input,
        clk		 =>     clk,
        cet      =>     '1',
        cep      =>     '1',
        sr_n     =>     pc_sr_n,
        pe_n     =>     pc_pe_n,
        tc       =>     pc_tc
    );
    
    -- P register/counter
    p : counter_74191
    generic map (
        WIDTH => 16,
        DELAY => 15 ns
    ) 
    port map (
        output   =>     ram_addr,
        input    =>     X"0000",
        clk		 =>     clk,
        cten_n	 =>     p_cten_n,
        du_n	 =>     p_du_n,
        load_n	 =>     p_load_n,
        rco_n    =>     p_rco_n,
        minmax   =>     p_minmax
    );
    
    -- A register/counter
    a : counter_74191
    generic map (
        WIDTH => 8,
        DELAY => 15 ns
    ) 
    port map (
        output   =>     a_output,
        input    =>     ram_data,
        clk		 =>     clk,
        cten_n	 =>     a_cten_n,
        du_n	 =>     a_du_n,
        load_n	 =>     a_load_n,
        rco_n    =>     a_rco_n,
        minmax   =>     a_minmax
    );

    -- buffer from A output to data bus
    ao : buff
    generic map (
        WIDTH => 8,
        DELAY => 8 ns
    )
    port map (
        output => ram_data,
        input => a_output,
        oe_n => ao_oe_n
    );
    
    -- buffer from PM output to control lines
    --co : buff
    --generic map (
    --    WIDTH => 16,
    --    DELAY => 8 ns
    --)
    --port map (
    --    output => co_out,
    --    input => pm_data(15 downto 0),
    --    oe_n => pm_data(0)
    --);
    
    -- 10 MHz main clock
    process
    begin
        --data <= (others => 'Z');
        clk <= '0';
        wait for 50 ns;
        clk <= '1';
        wait for 50 ns;
    end process;

    -- pull down all busses
    pm_addr <= (others => 'L');
    pm_data <= (others => 'L');
    ram_addr <= (others => 'L');
    ram_data <= (others => 'L');
    
    -- control logic
    co_out <= "0000000111111011" after 8 ns when (pm_data(0) = '1') else pm_data after 8 ns;
    pm_oe <= not reset;
    pc_pe_n <= (not pm_data(0)) or (not a_minmax);
    ram_wr_n <= co_out(1);
    ram_oe <= co_out(2);
    pc_sr_n <= not reset;
    p_cten_n <= co_out(3);
    p_load_n <= co_out(4);
    p_du_n <= co_out(5);
    a_cten_n <= co_out(6);
    a_load_n <= co_out(7);
    a_du_n <= co_out(5);
    ao_oe_n <= co_out(8);
   
    pc_input <= "0" & pm_data(15 downto 1);

    -- output control
    process(clk)
    begin
        if clk'event and clk = '1' and co_out(9) = '1' then
            report "OUT " & hstr(a_output);
        end if;
    end process;

    -- launch everything by pulsing reset
    process
    begin
        reset <= '0';
        wait for 10 ns;
        reset <= '1';
        wait for 50 ns;
        reset <= '0';
        wait for 10000 ms;
    end process;

end;
